/* 
 * The contents of this file are subject to the GrammarSoft Public
 * License Version 1.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.grammarsoft.com/GSPL
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is the VISL CG Parser.
 * 
 * The Initial Developer of the Original Code is Martin Carlsen, 
 * GrammarSoft, and SDU Odense University.  
 * Portions created by Martin Carlsen, Grammarsoft, and SDU Odense 
 * University are Copyright (C) 1999-2000 Martin Carlsen, GrammarSoft, 
 * and SDU Odense University.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License Version 2 or later (the
 * "GPL"), in which case the provisions of the GPL are applicable 
 * instead of those above.  If you wish to allow use of your 
 * version of this file only under the terms of the GPL and not to
 * allow others to use your version of this file under the GSPL,
 * indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by
 * the GPL.  If you do not delete the provisions above, a recipient
 * may use your version of this file under either the GSPL or the
 * GPL.
 */
#ifndef _CG_DATA_STRUCTURES_HH_
#define _CG_DATA_STRUCTURES_HH_

#include <list>
#include <map>
#include <hash_map>
#include <string>
#include <iostream>

const string * const empty_string = new string("");

class Set;
class SetReference;
class ListDeclaration;
class SetDeclaration;
class SetUnion;
class SetConcatenation;
class SetDifference;
class Position;
class Context;
class Correction;
class Constraint;
class Mapping;

/**
 * Visitor pattern. 
 */
class Visitor
{
public:
  virtual void visitSetReference(const SetReference * s) = 0;
  virtual void visitListDeclaration(const ListDeclaration * s) = 0;
  virtual void visitSetDeclaration(const SetDeclaration * s) = 0;
  virtual void visitSetConcatenation(const SetConcatenation * s) = 0;
  virtual void visitSetUnion(const SetUnion * s) = 0;
  virtual void visitSetDifference(const SetDifference * s) = 0;
  virtual void visitPosition(const Position * p) = 0;
  virtual void visitContext(const Context * c) = 0;
  virtual void visitCorrection(const Correction * c) = 0;
  virtual void visitConstraint(const Constraint * c) = 0;
  virtual void visitMapping(const Mapping * m) = 0;
};

/**
 * Hash function for strings
 */
template<> struct hash<string>
{
  size_t operator()(const string& s) const {
    return __stl_hash_string(s.c_str());
  }
};

template<class T> class taglist;

/**
 * Combined tag or a reading of a cohort.
 * For better performance, each tag is represented as an integer.
 * Use the TagRegister to translate between int and string
 * representations of tags.
 */
typedef taglist<int> Combined;

enum Scantype     { scan_asterisk, scan_doubleasterisk, scan_none };
enum CorrectionOp { op_append, op_substitute };
enum MappingOp    { op_add, op_map, op_replace };
enum ConstraintOp { op_remove, op_select };


template<class T>
class taglist : public list<T>
{
public:
  bool has(taglist<T> * reading) const;
  ostream& put(ostream&, bool noParen=true, 
	       hash_map<T,string>* tagmap=0) const;
};

/**
 * Register of tags, for converting between two representations of tags.
 * (string and int).
 *
 * The TagRegister is a Singleton:
 * There is only one instance of the TagRegister, which is available
 * through the instance() method.
 */
class TagRegister
{
private:
  string * marker;
  hash_map<string,int> * forward_map;
  hash_map<int,string> * reverse_map;
  /** special[i] = Whether the tag begins with the @-marker */
  hash_map<int,bool> * special; 
  
  static TagRegister * _instance;
  
  TagRegister()
    {
      forward_map = new hash_map<string,int>();
      reverse_map = new hash_map<int,string>();
      special = new hash_map<int,bool>();
      marker = new string("@");
    }
  
public:
  /**
   * Returns whether the tag begins with the @-marker
   */
  bool isSpecial(int tag) const 
    {
      hash_map<int,bool>::iterator i = special->find(tag);
      if(i != special->end()) return i->second;
      return false;
    }
  /**
   * Change the special marker from the default @
   */
  void setMarker(string & s)
    {
      delete marker;
      marker = new string(s);
    }
  hash_map<int,string> * getReverseMap() { return reverse_map; }
  
  /** Get the instance of the TagRegister */
  static TagRegister * instance();
  
  /** 
   * Add a tag to the register. It is safe to add a tag that already
   * exists in the register.
   * Returns the int representation of the tag.
   */
  int add(string & tag)
    {
      hash_map<string,int>::iterator i = forward_map->find(tag);
      if(i != forward_map->end()) return i->second;
      
      int u = forward_map->size();
      
      (*forward_map)[tag]=u;
      (*reverse_map)[u]=tag;
      (*special)[u] = tag.find(*marker)==0;
      return u;
    }
  
  unsigned int size() const { return reverse_map->size(); }
  
  /**
   * Translate an int representation of a tag to its string representation.
   * Returns the empty string if the tag is not in the register.
   */
  string get(int tag)
    {
      hash_map<int,string>::iterator i = reverse_map->find(tag);
      if(i != reverse_map->end()) return i->second;
      return "";
    }
  /**
   * Translate a string representation of a tag to its int representation.
   * Returns -1 if the tag is not in the register.
   */
  int get(const string & tag)
    {
      hash_map<string,int>::iterator i = forward_map->find(tag);
      if(i != forward_map->end()) return i->second;
      return -1;
    }
};

/**
 * A word and its possible readings.
 * The input to the compiler is a list of Cohorts.
 */
class Cohort
{
private:
  string wordform;
  list<Combined*>* readings;
  /** 
   * The wordform represented as a combined tag.
   * In minimal trace mode, this may include additional trace information.
   * The wordform itself is always the first "tag" in the list.
   */
  Combined * wf_reading;
  
public:
  list<Combined*>* getReadings() const { return readings; }
  const string & getWordform() const { return wordform; }
  Combined * getWordformReading() const { return wf_reading; }
  
  Cohort(const string & wf = *empty_string, list<Combined*>* reads=0)
    : wordform(wf), readings(reads)
    {
      wf_reading = new Combined();
      wf_reading->push_back(TagRegister::instance()->add(wordform));
    }
  
  ~Cohort();
  
  friend ostream& operator<< (ostream& os, const Cohort& c);
};

#include "Set.hh"

/**
 * position ::= scantype? absolute? offset careful?
 * 
 * e.g.:
 * **-1 
 * where:
 *   ** is the scantype
 *   -1 is the offset
 * @1C
 * where:
 *   @ indicates absolute position in the sentence
 *   1 is the offset
 *   C indicates careful mode
 */
class Position
{
  friend class Visitor;
private:
  Scantype scantype;
  bool absolute;
  int offset;
  bool careful;
  
public:
  Scantype getScantype() const { return scantype; }
  int getOffset() const { return offset; }
  bool isAbsolute() const { return absolute; }
  bool isCareful() const { return careful; }
  void accept(Visitor * v) { v->visitPosition(this); }
  
  Position(Scantype type=scan_none, bool absol=false, int off=0, bool care=false)
    : scantype(type), absolute(absol), offset(off), careful(care)
    {
    }
  
  friend ostream& operator<< (ostream& os, const Position& c);
#ifdef NEVER
  ostream& gen(ostream& os) const;
#endif
};

/**
 * context ::= not? position set barrier? link?
 *
 * e.g.: 
 * (NOT -1C DET) 
 * where:
 *   NOT indicates negative context
 *   -1C is the position 
 *   DET is the set to test readings against
 * (*-1 N BARRIER DET LINK -1 @SUBJ)
 * where:
 *   *-1 is the position
 *   N is the set to test readings against
 *   DET is the barier
 *   -1 @SUBJ is the linked context
 */
class Context
{
  friend class Visitor;
private:
  bool negative;
  Position * position;
  Set * testset;
  Set * barrier;
  Context * link;
  
public:
  virtual bool isNegative() const { return negative; }
  virtual Position * getPosition() const { return position; }
  virtual Set * getTest() const { return testset; }
  virtual Set * getBarrier() const { return barrier; }
  virtual Context * getLink() const { return link; }
  void accept(Visitor * v) { v->visitContext(this); }
  
  virtual bool matchCohort(vector<Cohort*>* cohorts, unsigned int pos, bool asLink=false) const;
  
  Context(bool neg=false, Position* pos=0, Set* test=0, Set* barr=0, Context* lnk=0)
    : negative(neg), position(pos), testset(test), barrier(barr), link(lnk)
    {
    }

  friend ostream& operator<< (ostream& os, const Context& c);
#ifdef NEVER
  ostream& gen(ostream& os) const;
#endif
};

/**
 * correction ::= wordform? APPEND tag contextualtests*
 *              | wordform? SUBSTITUTE tag tag target contextualtests*
 */
class Correction
{
  friend class Visitor;
private:
  string wordform;
  CorrectionOp op;
  Combined * removal;
  Combined * insertion;
  Set * target;
  list<Context*>* tests;
  int line;
  
public:
  string getWordform() const { return wordform; }
  CorrectionOp getOp() const { return op; }
  Combined * getRemoval() const { return removal; }
  Combined * getInsertion() const { return insertion; }
  Set * getTarget() const { return target; }
  list<Context*>* getTests() const { return tests; }
  int getLine() const { return line; }
  void accept(Visitor * v) { v->visitCorrection(this); }
  
  bool correct(vector<Cohort*>* cohorts, int pos, list<Combined*>::iterator ri) const;
  
  Correction(string wf="", CorrectionOp correctionop=op_append, Combined* remove=0, Combined* insert=0, Set* targ=0, list<Context*>* test=0, int l=0)
    : wordform(wf), op(correctionop), removal(remove), insertion(insert), target(targ), tests(test), line(l)
    {
    }
  
  friend ostream& operator<< (ostream& os, const Correction& c);
#ifdef NEVER
  ostream& gen(ostream& os) const;
#endif
};

/**
 * mapping ::= map-operator tag+ target contextualtests*
 */
class Mapping
{
  friend class Visitor;
private:
  MappingOp op;
  Combined * insertion;
  Set * target;
  list<Context*>* tests;
  int line;
  
public:
  MappingOp getOp() const { return op; }
  Combined * getInsertion() const { return insertion; }
  Set * getTarget() const { return target; }
  list<Context*>* getTests() const { return tests; }
  int getLine() const { return line; }
  void accept(Visitor * v) { v->visitMapping(this); }
  
  bool map(vector<Cohort*>* cohorts, int pos, Combined * reading) const;
  
  Mapping(MappingOp mapop=op_map, Combined* insert=0, Set* targ=0, list<Context*>* test=0, int l=0)
    : op(mapop), insertion(insert), target(targ), tests(test), line(l)
    {
    }
  
  friend ostream& operator<< (ostream& os, const Mapping& c);
#ifdef NEVER
  ostream& gen(ostream& os) const;
#endif
};

/**
 * constraint ::= wordform? constraint-operator target? contextualtests*
 */
class Constraint
{
  friend class Visitor;
private:
  string wordform;
  ConstraintOp op;
  Set * target;
  list<Context*>* tests;
  int line;
  
public:
  string getWordform() const { return wordform; }
  ConstraintOp getOp() const { return op; }
  Set * getTarget() const { return target; }
  list<Context*>* getTests() const { return tests; }
  int getLine() const { return line; }
  void accept(Visitor * v) { v->visitConstraint(this); }
  
  bool disambiguate(vector<Cohort*>* cohorts, int pos, list<Combined*>::iterator ri) const;
  
  Constraint(string wf="", ConstraintOp cop=op_remove, Set* targ=0, list<Context*>* test=0, int l=0)
    : wordform(wf), op(cop), target(targ), tests(test), line(l)
    {
    }
  
  friend ostream& operator<< (ostream& os, const Constraint& c);
#ifdef NEVER
  ostream& gen(ostream& os) const;
#endif
};

ostream& operator<< (ostream& os, const Set& s);
ostream& operator<< (ostream& os, const Cohort& c);
ostream& operator<< (ostream& os, const Position& s);
ostream& operator<< (ostream& os, const Context& s);
ostream& operator<< (ostream& os, const Constraint& s);
ostream& operator<< (ostream& os, const Mapping& s);

#endif
