/* 
 * The contents of this file are subject to the GrammarSoft Public
 * License Version 1.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.grammarsoft.com/GSPL
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is the VISL CG Parser.
 * 
 * The Initial Developer of the Original Code is Martin Carlsen, 
 * GrammarSoft, and SDU Odense University.  
 * Portions created by Martin Carlsen, Grammarsoft, and SDU Odense 
 * University are Copyright (C) 1999-2000 Martin Carlsen, GrammarSoft, 
 * and SDU Odense University.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * Alternatively, the contents of this file may be used under the
 * terms of the GNU General Public License Version 2 or later (the
 * "GPL"), in which case the provisions of the GPL are applicable 
 * instead of those above.  If you wish to allow use of your 
 * version of this file only under the terms of the GPL and not to
 * allow others to use your version of this file under the GSPL,
 * indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by
 * the GPL.  If you do not delete the provisions above, a recipient
 * may use your version of this file under either the GSPL or the
 * GPL.
 */
#include <map>
#include <string>
#include <vector>
#include <fstream>
#include <streambuf.h>

//#include <mcheck.h>

#include "cg.data_structures.hh"
#include "cohorts.hh"
#include "CGCompiler.hh"
#include "Options.hh"
#include "Debug.hh"

#include "Generator.hh"

extern ostream * os_bin;

CGCompiler * buildCompiler(istream* yyin = 0, Generator* gen = 0 );
#ifndef GENERATOR
Generator * createGenerator() { return 0; }
void generateSourceFiles(string bin_file, string cc_file) {}
#endif

Options * buildOptions()
{
  map<string,string>* opts = new map<string,string>();
  
  (*opts)["grammar"] =
    "--grammar=rulesfile or --grammar rulesfile\n"
    "        Specifies which rules file to use.\n"
    "        This option must be specified if any parsing is to be done.\n";
  (*opts)["prefix"] =
    "--prefix=marker or --prefix marker\n"
    "        This changes the special marker for ambiguity and mappings.\n"
    "        The marker may be more than one character.\n"
    "        The default marker is the '@' character.\n";
  (*opts)["sections"] =
    "--sections=n or --sections n\n"
    "        This sets the number (n) of CONSTRAINT sections that are applied.\n"
    "        The default is to apply all sections.\n";
  (*opts)["check-only"] =
    "--check-only\n" 
    "        Test the syntax of the rules file and exit before performing any\n"
    "        mappings or disambiguations.\n";
  (*opts)["quiet"] =
    "--quiet\n" 
    "        Do not output any messages. This option overrides any debug and\n"
    "        trace options.\n";
  (*opts)["no-mappings"] =
    "--no-mappings\n" 
    "        Do not apply any mapping rules. This corresponds to CG-2's dis module\n";
  (*opts)["no-corrections"] =
    "--no-corrections\n" 
    "        Do not apply any correction rules.\n";
  (*opts)["unsafe"] =
    "--unsafe\n" 
    "        Allow the compiler to remove the last reading of a cohort.\n";
  (*opts)["reorder"] =
    "--reorder\n"
    "        Attempt to reorder rules in a way similar to CG-2, using\n"
    "        PREFERRED-TARGETS and rules of thumb.\n"
    "        [ The default is to apply rules in the strict order they appear\n"
    "        in the rules file and ignore PREFERRED-TARGETS ]\n";
  (*opts)["help"] =
    "--help\n"
    "        Print this text.\n";
#ifndef BINARY
  (*opts)["trace-mappings"] =
    "--trace-mappings\n"
    "        Trace which mappings were applied and the order of application.\n";
  (*opts)["trace-constraints"] =
    "--trace-constraints\n"
    "        Trace which constraints were applied and the order of application.\n";
  (*opts)["trace"] =
    "--trace\n" 
    "        Equivalent to --trace-mappings --trace-constraints.\n";
  (*opts)["visldebug"] =
    "--visldebug\n" 
    "        Run compiler in rules debugging mode.\n";
  (*opts)["cg2-debug"] =
    "--debug\n" 
    "        Run compiler in CG2-like rules debugging mode.\n";
  (*opts)["verbosity"] =
    "--verbosity=level or --verbosity level\n"
    "        This option causes extra or less information to be printed in trace\n"
    "        mode. Valid levels are:\n"
    "        'minimal' - Line numbers of applied rules are attached to the output\n"
    "                    cohorts. The output in minimal mode is similar to that\n"
    "                    of CG-2's mdis.trace module.\n"
    "        'normal'  - Instead of attaching line numbers to readings, each rule\n"
    "                    (the full specification - not just the line number) is\n"
    "                    printed each time it is applied.\n"
    "                    [ 'normal' is the DEFAULT trace mode verbosity.]\n"
    "        'verbose' - As normal mode, but the cohort to which each rule was\n"
    "                    applied is printed as well. This will also cause vislcg\n"
    "                    to print all sets, mappings and constraints before\n"
    "                    processing any input.\n";
  (*opts)["minimal"] =
    "--minimal\n"
    "        Equivalent to --verbosity=minimal\n";
  (*opts)["verbose"] =
    "--verbose\n"
    "        Equivalent to --verbosity=verbose\n";
#endif
#ifdef GENERATOR
  (*opts)["generate"] =
    "--generate=outputfile or --generate outputfile\n"
    "        Generate binary rules file for use in closed grammars.\n"
    "        If outputfile is omitted, the binary file will be written\n"
    "        to stdout\n";
#endif
  return new Options(opts);
}

int main(int argc, char * argv[])
{ 
  string command(argv[0]);
  string usage = "Usage: " + command + " [OPTION]...\n"
    "(\"" + command + " --help\" for help on options)\n";
  
  Options * options = buildOptions();
  options->parse(argc,argv);
  
  // --help
  if(options->getSetting("help","no") == "")
    {
      cout << usage;
      cout << options->describe() << "\n";
      return 0;
    }
  
  // Open rule file (option --grammar)
  string grammar = options->getSetting("grammar","");
  if(grammar=="")
    {
      qerr << "No rule file specified (use option --grammar)\n";
      qerr << usage;
      return 1;
    }
  char * p = new char[grammar.length()+1];
  grammar.copy(p,string::npos);
  p[grammar.length()] = 0;
  ifstream * yyin = new ifstream(p);
  if(!(*yyin))
    {
      qerr << "Could not open rules file " << grammar <<"\n";
      return 1;
    }
  
  // initialize debug options
  Debug * debug = 0;
#ifdef BINARY
  debug = Debug::instance();
#else
  debug = Debug::init(options);
#endif

  // other options
  // --reorder
  bool no_reordering = options->getSetting("reorder","no") != "";
  // --unsafe
  bool allow_unsafe = options->getSetting("unsafe","no") == "";
  // --no-mappings
  bool apply_mappings = 
    options->getSetting("no-mapping","no") != "" &&
    options->getSetting("no-mappings","no") != "";
  bool apply_corrections = 
    options->getSetting("no-correction","no") != "" &&
    options->getSetting("no-corrections","no") != "";
  // --prefix
  string prefix = options->getSetting("prefix","@");
  TagRegister::instance()->setMarker(prefix);
  // --sections
  int nSections = -1;
  string sections = options->getSetting("sections","");
  if(sections!="") nSections = atoi(sections.c_str());
  
  // --generate
  Generator * gen = 0;
#ifdef GENERATOR
  bool generate = options->getSetting("generate","//\\\\") != "//\\\\";
  string out_bin = options->getSetting("generate","");
  if(generate) 
    {
      debug->check_only = true;
      gen = createGenerator();  
    }
#endif
  
  // Parse rules.
  CGCompiler * compiler = buildCompiler(yyin, gen);
  
  if(!no_reordering) compiler->reorderRules();
  
#ifdef GENERATOR
  if(generate)
    {
      gen->setCompiler(compiler);
      generateSourceFiles(out_bin);
    }
#endif
  
  if(debug->check_only) return 0;
  
  // Compile input.
  // One sentence at a time.
  vector<Cohort*> * cohorts = 0;

  while( (cohorts = readCohorts(compiler->getDelimiters())) )
    {
      if(apply_corrections)
	{
	  // Expand readings with multiple @-tags 
	  // to multiple readings with each one @-tag
	  CGCompiler::expandCohorts(cohorts);
	  // Apply correction rules
	  compiler->applyCorrections(cohorts);
	}
      if(apply_mappings)
	{
	  // Expand readings with multiple @-tags 
	  // to multiple readings with each one @-tag
	  CGCompiler::expandCohorts(cohorts);
	  // Apply mapping rules
	  compiler->applyMapping(cohorts);
	}
      // Expand readings with multiple @-tags 
      // to multiple readings with each one @-tag
      // (again, because the mappings may have introduced further ambiguities)
      
      CGCompiler::expandCohorts(cohorts);

      // Apply constraint rules
      compiler->applyConstraints(cohorts, allow_unsafe, nSections);
      
      // Print result
      vector<Cohort*>::iterator i = cohorts->begin();
      delete *i;
      ++i; // skip the first (>>>) cohort.
      for(vector<Cohort*>::iterator 
	    z=cohorts->end();
	  i!=z; ++i)
	{
	  cout << (**i);
	  delete *i;
	}
      
      delete cohorts;
    }
  
  return 0;
}
